@@ -33,8 +33,8 @@ typedef struct Node_ {
     dev_t device;
 } Node;
 
-Node *nodes;
-size_t nodes_size;
+static Node *nodes;
+static size_t nodes_size;
 
 static void oom(void)
 {
@@ -65,19 +65,20 @@ static int init_tz(void)
     return 0;
 }
 
-static int traversal(const char * const s)
+static int traverse(const char * const s)
 {
     DIR *d;
     struct dirent *de;
     struct stat st;
     size_t slen;
     Node *nodes_pnt = nodes;
-    size_t nodes_cnt = nodes_size;
+    Node *found_node;
+    size_t nodes_sizeleft = nodes_size;
     int fd;
-    char *buf = NULL;
-    size_t sizeof_buf = (size_t) 0U;
+    char *path = NULL;
+    size_t sizeof_path = (size_t) 0U;
 
-    if ((fd = open(s, O_RDONLY | O_DIRECTORY)) == -1) {
+    if ((fd = open(s, O_RDONLY | O_DIRECTORY | O_NONBLOCK)) == -1) {
         if (errno != EACCES) {
             return -1;
         }
@@ -87,7 +88,7 @@ static int traversal(const char * const s)
         }
         (void) fchmod(fd, st.st_mode | 0500);
         close(fd);
-        if ((fd = open(s, O_RDONLY | O_DIRECTORY)) == -1) {
+        if ((fd = open(s, O_RDONLY | O_DIRECTORY | O_NONBLOCK)) == -1) {
             return -1;
         }
     }
@@ -99,12 +100,12 @@ static int traversal(const char * const s)
         (void) fchmod(fd, (mode_t) (st.st_mode | 0500)); /* if it fails, try anyway */
     }
     close(fd);
-    while (nodes_cnt > (size_t) 0U) {
+    while (nodes_sizeleft > (size_t) 0U) {
         if (nodes_pnt->inode == st.st_ino && nodes_pnt->device == st.st_dev) {
             return -1;
         }
         nodes_pnt++;
-        nodes_cnt -= sizeof *nodes_pnt;
+        nodes_sizeleft -= sizeof *nodes_pnt;
     }
     if (nodes == NULL) {
         if ((nodes = malloc(sizeof *nodes)) == NULL) {
@@ -119,20 +120,17 @@ static int traversal(const char * const s)
         }
         nodes = new_nodes;
     }
-    {
-        Node * const node =
-            (Node *) (void *) (((unsigned char *) nodes) + nodes_size);
-
-        node->inode = st.st_ino;
-        node->device = st.st_dev;
-    }
+    found_node =
+        (Node *) (void *) (((unsigned char *) nodes) + nodes_size);
+    found_node->inode = st.st_ino;
+    found_node->device = st.st_dev;
     nodes_size += sizeof *nodes_pnt;
     if ((d = opendir(s)) == NULL) {
         return -1;
     }
     slen = strlen(s) + (size_t) 2U;
     while ((de = readdir(d)) != NULL) {
-        size_t wanted_sizeof_buf;
+        size_t wanted_sizeof_path;
 
         if ((de->d_name[0] == '.' && de->d_name[1] == 0) ||
             (de->d_name[0] == '.' && de->d_name[1] == '.' &&
@@ -142,17 +140,17 @@ static int traversal(const char * const s)
         if (strcmp(de->d_name, QUOTA_FILE) == 0) {
             continue;
         }
-        wanted_sizeof_buf = slen + strlen(de->d_name);
-        if (wanted_sizeof_buf > sizeof_buf) {
-            if ((buf = realloc(buf, wanted_sizeof_buf)) == NULL) {
+        wanted_sizeof_path = slen + strlen(de->d_name);
+        if (wanted_sizeof_path > sizeof_path) {
+            if ((path = realloc(path, wanted_sizeof_path)) == NULL) {
                 oom();
             }
-            sizeof_buf = wanted_sizeof_buf;
+            sizeof_path = wanted_sizeof_path;
         }
-        snprintf(buf, sizeof_buf, "%s/%s", s, de->d_name);
-        if (stat(buf, &st) == 0) {
+        snprintf(path, sizeof_path, "%s/%s", s, de->d_name);
+        if (stat(path, &st) == 0) {
             if (S_ISDIR(st.st_mode)) {
-                if (traversal(buf) == 0) {
+                if (traverse(path) == 0) {
                     total_files++;
                 }
             } else if (S_ISREG(st.st_mode)) {
@@ -161,7 +159,7 @@ static int traversal(const char * const s)
             }
         }
     }
-    free(buf);
+    free(path);
     closedir(d);
 
     return 0;
@@ -232,7 +230,7 @@ static int writequota(const char * const quota_file)
     char buf[84];
     const char *bufpnt = buf;
 
-    if ((fd = open("/", O_RDONLY | O_DIRECTORY)) == -1) {
+    if ((fd = open("/", O_RDONLY | O_DIRECTORY | O_NONBLOCK)) == -1) {
         return -1;
     }
     if (fstat(fd, &st) != 0 || !S_ISDIR(st.st_mode)) {
@@ -338,10 +336,6 @@ int main(int argc, char *argv[])
             {
                 struct passwd *pw;
 
-                if (uid > (uid_t) 0) {
-                    fprintf(stderr, "You already gave an uid\n");
-                    return -1;
-                }
                 if ((pw = getpwnam(optarg)) != NULL) {
                     uid = pw->pw_uid;
                     if (gid == (gid_t) 0) {
@@ -389,7 +383,7 @@ int main(int argc, char *argv[])
                 startpath, strerror(errno));
         return -3;
     }
-    if (traversal(isroot != 0 ? "/" : "./") < 0) {
+    if (traverse(isroot != 0 ? "/" : "./") < 0) {
         fprintf(stderr, "Unable to traverse [%s]: [%s]\n",
                 startpath, strerror(errno));
         free(nodes);
